using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;

using Str = System.Collections.Generic.IEnumerable<char>;
using OpPriority = System.SByte;

namespace MathParser
{
    public class FctParser<T>
    {
        public T Parse(Str function, T x)
        {
            sFunction = function;
            xValue = x;

            Functionator = function.GetEnumerator();

            char c = '\0';
            T y = Expression(ref c);
            if(c != '\0')
                throw new FctException("Unexpected character at end of expression", c);

            return y;
        }

        public T Parse(ref char c, ref IEnumerator<char> functionator, T x)
        {
            xValue = x;
            Functionator = functionator;

            T y = Calculate(ref c, 1);

            functionator = Functionator;

            return y;
        }

        public static CultureInfo Culture = new CultureInfo(String.Empty);

        protected char Dot
        {
            get { return Culture.NumberFormat.NumberDecimalSeparator[0]; }
        }

        protected bool IgnoreCase = true;

        // Variable

        protected string X = "x";

        // Bracket

        protected struct Bracket
        {
            public Bracket(char bra, char ket)
            {
                this.bra = bra;
                this.ket = ket;
            }

            public char bra, ket;

            public static Bracket Default = new Bracket('(', ')');
        }

        // Function

        protected delegate T Fct(T x);

        protected class Function
        {
            public Function(string sFct, Fct fct)
            {
                this.sFct = sFct;
                this.fct = fct;
            }

            public string sFct;
            public Fct fct;
        }

        protected IEnumerable<Function> Functions;
        protected Bracket FunctionBracket = Bracket.Default;

        // Operator

        protected delegate T Op(T x, T y);

        protected enum OpType
        {
            Unary, Binary
        }

        protected class Operator
        {
            public Operator(string sOp, Op op, OpType type, OpPriority pri)
            {
                this.sOp = sOp;
                this.op = op;
                this.type = type;
                this.priority = pri;
            }

            public Operator(char cOp, Op op, OpType type, OpPriority pri)
                : this(new string(cOp, 1), op, type, pri)
            {
            }

            public string sOp;
            public Op op;
            public OpType type;
            public OpPriority priority;
        }

        protected OpPriority MaxOpPriority;
        protected IEnumerable<Operator> Operators;
        protected Bracket OperatorBracket = Bracket.Default;

        // Constant

        protected class Constant
        {
            public Constant(string sConst, T value)
            {
                this.sConst = sConst;
                this.value = value;
            }

            public string sConst;
            public T value;
        }

        protected IEnumerable<Constant> Constants;

        // Fields

        protected Str sFunction;
        protected T xValue;

        protected IEnumerator<char> Functionator;

        // Methods

        protected char GetChar()
        {
            char c;
            do
            {
                c = GetNextChar();
            }
            while(c != '\0' && Char.IsWhiteSpace(c));

            return c;
        }

        protected char GetNextChar()
        {
            return Functionator.MoveNext()? Functionator.Current : '\0';
        }

        // Algorithmus
        /*
        Funktionstext zeichenweise parsen:
        Terminalsymbole sind Zahlen, Variablen, Konstanten u. Funktionen
        Funktionen werden erst nach Parsen des Parameters ausgefhrt (Prfix-Notation)
        Unre Operatoren knnen vor einem Terminalsymbol stehen (Prfix-Notation)
        Binre Operatoren stehen zwischen zwei Terminalsymbolen (Infix-Notation)
        Je nach Prioritt werden die Operationen ausgefhrt,
        bei gleicher Prioritt erfolgt die Auswertung von links nach rechts
        Operatoren-Klammern haben die hchste Prioritt
        */

        T Expression(ref char c)
        {
            c = GetChar();
            T y = Calculate(ref c, 1);

            return y;
        }

        T Calculate(ref char c, OpPriority op_pri)
        {
            T y = Calc(ref c, op_pri);

            Operator op;
            while((op = GetOperator(ref c, OpType.Binary, op_pri)) != null)
            {
                T y2 = Calc(ref c, op_pri);

                y = op.op(y, y2);  // Binre Operation ausfhren
            }

            return y;
        }

        T Calc(ref char c, OpPriority op_pri)
        {
            T y;

            if(op_pri < MaxOpPriority)
                y = Calculate(ref c, (OpPriority)(op_pri + 1));
            else
                y = GetTerminal(ref c);

            return y;
        }

        T GetTerminal(ref char c)
        {
            T y;
            Operator op = GetOperator(ref c, OpType.Unary, 0);

            if(c == OperatorBracket.bra)
            {
                y = Expression(ref c);
                if(c != OperatorBracket.ket)
                    throw new FctException("Missing bracket", OperatorBracket.ket);
                c = GetChar();
            }
            else if(IsNumberChar(c))
            {
                y = GetNumber(ref c);
            }
            else if(IsIdentifierChar(c, true))
            {
                string sId = GetIdentifier(ref c);
                // Variable, Konstante oder Funktion
                if(IsVariable(sId))
                    y = xValue;
                else
                {
                    Constant ct = IsConstant(sId);
                    if(ct != null)
                        y = ct.value;
                    else
                    {
                        Function f = IsFunction(sId);
                        if(f != null)
                        {
                            if(c != FunctionBracket.bra)
                                throw new FctException("Missing function bracket", FunctionBracket.bra);
                            T x = Expression(ref c);
                            if(c != FunctionBracket.ket)
                                throw new FctException("Missing function bracket", FunctionBracket.ket);
                            c = GetChar();
                            y = f.fct(x); // Funktion ausfhren
                        }
                        else
                            throw new FctException("Unknown identifier", sId);
                    }
                }
            }
            else
                throw new FctException("Number or identifier expected instead of", c);

            if(op != null)
                y = op.op(default(T), y); // Unre Operation ausfhren

            return y;
        }

        T GetNumber(ref char c)
        {
            T y = ParseNumber(ref c);

            if(Char.IsWhiteSpace(c))
                c = GetChar();

            return y;
        }

        protected virtual bool IsNumberChar(char c)
        {
            return Char.IsDigit(c) || c == Dot;
        }

        protected virtual T ParseNumber(ref char c)
        {
            double y = ParseDouble(ref c);

            return (T)Convert.ChangeType(y, typeof(T));
        }

        protected double ParseDouble(ref char c)
        {
            StringBuilder s = new StringBuilder(32);

            bool bDot = (c == Dot);

            do
            {
                s.Append(c);
                c = GetNextChar();
                if(c == Dot)
                {
                    if(bDot)
                        break;
                    else
                        bDot = true;
                }
            }
            while(Char.IsDigit(c) || c == Dot);

            if(Char.ToUpper(c) == 'E')
            {
                s.Append(c);
                c = GetNextChar();
                if(c == '+' || c == '-')
                {
                    s.Append(c);
                    c = GetNextChar();
                }

                while(Char.IsDigit(c))
                {
                    s.Append(c);
                    c = GetNextChar();
                }
            }

            return Double.Parse(s.ToString(), Culture.NumberFormat);
        }

        protected int ParseInt(ref char c)
        {
            int y = 0;

            while(Char.IsDigit(c))
            {
                y = 10 * y + (c - '0');

                c = GetNextChar();
            }

            return y;
        }

        delegate bool IsTokenChar(char c);

        string GetToken(ref char c, IsTokenChar IsTokenChar)
        {
            string s = String.Empty;

            s += c;
            while(true)
            {
                c = GetNextChar();
                if(!IsTokenChar(c))
                    break;

                s += c;
            }

            if(Char.IsWhiteSpace(c))
                c = GetChar();

            return s;
        }

        Operator GetOperator(ref char c, OpType op_type, OpPriority op_pri)
        {
            if(!IsOpChar(c))
                return null;

            // save current Functionator and character
            IEnumerator<char> functionator = (IEnumerator<char>)((CharEnumerator)Functionator).Clone();
            char cOp = c;

            string sOp = GetOperator(ref c);
            Operator op = IsOperator(sOp, op_type, op_pri);
            if(op == null)
            {
                Functionator = functionator;
                c = cOp;
            }

            return op;
        }

        string GetOperator(ref char c)
        {
            return GetToken(ref c, IsOpChar);
        }

        protected bool IsBracketChar(Bracket bracket, char c)
        {
            return (c == bracket.bra) || (c == bracket.ket);
        }

        protected bool IsOperatorBracket(char c)
        {
            return IsBracketChar(OperatorBracket, c);
        }

        protected bool IsOpChar(char c)
        {
            return !IsOperatorBracket(c) && IsOperatorChar(c);
        }

        protected virtual bool IsOperatorChar(char c)
        {
            return (c != Dot) && (Char.IsPunctuation(c) || Char.IsSymbol(c));
        }

        string GetIdentifier(ref char c)
        {
            return GetToken(ref c, delegate(char cId) { return IsIdentifierChar(cId, false); });
        }

        protected virtual bool IsIdentifierChar(char c, bool bFirstChar)
        {
            return Char.IsLetter(c);
        }

        bool IsEqual(string sX, string sY)
        {
            return String.Compare(sX, sY, IgnoreCase, Culture) == 0;
        }

        bool IsVariable(string sVar)
        {
            return IsEqual(sVar, X);
        }

        Function IsFunction(string sFct)
        {
            foreach(Function f in Functions)
                if(IsEqual(sFct, f.sFct))
                    return f;

            return null;
        }

        Operator IsOperator(string sOp, OpType op_type, OpPriority op_pri)
        {
            foreach(Operator op in Operators)
                if(op_type == op.type && sOp == op.sOp)
                    if(op_pri == 0 || op_pri == op.priority)
                        return op;

            return null;
        }

        Constant IsConstant(string sConst)
        {
            foreach(Constant c in Constants)
                if(IsEqual(sConst, c.sConst))
                    return c;

            return null;
        }
    }

    public class FctException : Exception
    {
        public FctException(string sMessage)
            : base(sMessage)
        {
        }

        public FctException(string sMessage, object parameter)
            : base(sMessage + ": " + parameter.ToString())
        {
        }
    }
}
